/*--------------------------------------------------------------------------*\
 * pl_setcontlabelformatCmd
 *
 * Processes pl_setcontlabelformat Tcl command.
\*--------------------------------------------------------------------------*/

static int
pl_setcontlabelformatCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT lexp;
    PLINT sigprec;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pl_setcontlabelformat lexp sigprec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pl_setcontlabelformat lexp sigprec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    lexp = atoi(argv[1+0]);
    sigprec = atoi(argv[1+1]);

    pl_setcontlabelformat ( lexp, sigprec );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pl_setcontlabelparamCmd
 *
 * Processes pl_setcontlabelparam Tcl command.
\*--------------------------------------------------------------------------*/

static int
pl_setcontlabelparamCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT offset;
    PLFLT size;
    PLFLT spacing;
    PLINT active;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pl_setcontlabelparam offset size spacing active", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pl_setcontlabelparam offset size spacing active", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    offset = atof(argv[1+0]);
    size = atof(argv[1+1]);
    spacing = atof(argv[1+2]);
    active = atoi(argv[1+3]);

    pl_setcontlabelparam ( offset, size, spacing, active );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pladvCmd
 *
 * Processes pladv Tcl command.
\*--------------------------------------------------------------------------*/

static int
pladvCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT page = 0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pladv page", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 1 && (argc < (1 + 1 - 1))) ||
         (!0 && !1 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pladv page", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if (argc > 0+1) {    
    page = atoi(argv[1+0]);
    }

    pladv ( page );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plarcCmd
 *
 * Processes plarc Tcl command.
\*--------------------------------------------------------------------------*/

static int
plarcCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT x;
    PLFLT y;
    PLFLT a;
    PLFLT b;
    PLFLT angle1;
    PLFLT angle2;
    PLFLT rotate;
    PLINT fill;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plarc x y a b angle1 angle2 rotate fill", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 8 - 0))) ||
         (!0 && !0 && (argc != (8 + 1))) ||
         ( 0 && (argc != 1) && (argc != (8 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plarc x y a b angle1 angle2 rotate fill", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    x = atof(argv[1+0]);
    y = atof(argv[1+1]);
    a = atof(argv[1+2]);
    b = atof(argv[1+3]);
    angle1 = atof(argv[1+4]);
    angle2 = atof(argv[1+5]);
    rotate = atof(argv[1+6]);
    fill = atoi(argv[1+7]);

    plarc ( x, y, a, b, angle1, angle2, rotate, fill );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plaxesCmd
 *
 * Processes plaxes Tcl command.
\*--------------------------------------------------------------------------*/

static int
plaxesCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT x0;
    PLFLT y0;
    const char *xopt;
    PLFLT xtick;
    PLINT nxsub;
    const char *yopt;
    PLFLT ytick;
    PLINT nysub;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plaxes x0 y0 xopt xtick nxsub yopt ytick nysub", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 8 - 0))) ||
         (!0 && !0 && (argc != (8 + 1))) ||
         ( 0 && (argc != 1) && (argc != (8 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plaxes x0 y0 xopt xtick nxsub yopt ytick nysub", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    x0 = atof(argv[1+0]);
    y0 = atof(argv[1+1]);
    xopt = argv[1+2];
    xtick = atof(argv[1+3]);
    nxsub = atoi(argv[1+4]);
    yopt = argv[1+5];
    ytick = atof(argv[1+6]);
    nysub = atoi(argv[1+7]);

    plaxes ( x0, y0, xopt, xtick, nxsub, yopt, ytick, nysub );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plbinCmd
 *
 * Processes plbin Tcl command.
\*--------------------------------------------------------------------------*/

static int
plbinCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT nbin;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLINT center;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plbin nbin x y center", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plbin nbin x y center", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    nbin = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    center = atoi(argv[1+3]);

    plbin ( nbin, x, y, center );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plbopCmd
 *
 * Processes plbop Tcl command.
\*--------------------------------------------------------------------------*/

static int
plbopCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plbop ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plbop ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plbop (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plboxCmd
 *
 * Processes plbox Tcl command.
\*--------------------------------------------------------------------------*/

static int
plboxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *xopt;
    PLFLT xtick;
    PLINT nxsub;
    const char *yopt;
    PLFLT ytick;
    PLINT nysub;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plbox xopt xtick nxsub yopt ytick nysub", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plbox xopt xtick nxsub yopt ytick nysub", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xopt = argv[1+0];
    xtick = atof(argv[1+1]);
    nxsub = atoi(argv[1+2]);
    yopt = argv[1+3];
    ytick = atof(argv[1+4]);
    nysub = atoi(argv[1+5]);

    plbox ( xopt, xtick, nxsub, yopt, ytick, nysub );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plbox3Cmd
 *
 * Processes plbox3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plbox3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *xopt;
    const char *xlabel;
    PLFLT xtick;
    PLINT nsubx;
    const char *yopt;
    const char *ylabel;
    PLFLT ytick;
    PLINT nsuby;
    const char *zopt;
    const char *zlabel;
    PLFLT ztick;
    PLINT nsubz;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plbox3 xopt xlabel xtick nsubx yopt ylabel ytick nsuby zopt zlabel ztick nsubz", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 12 - 0))) ||
         (!0 && !0 && (argc != (12 + 1))) ||
         ( 0 && (argc != 1) && (argc != (12 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plbox3 xopt xlabel xtick nsubx yopt ylabel ytick nsuby zopt zlabel ztick nsubz", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xopt = argv[1+0];
    xlabel = argv[1+1];
    xtick = atof(argv[1+2]);
    nsubx = atoi(argv[1+3]);
    yopt = argv[1+4];
    ylabel = argv[1+5];
    ytick = atof(argv[1+6]);
    nsuby = atoi(argv[1+7]);
    zopt = argv[1+8];
    zlabel = argv[1+9];
    ztick = atof(argv[1+10]);
    nsubz = atoi(argv[1+11]);

    plbox3 ( xopt, xlabel, xtick, nsubx, yopt, ylabel, ytick, nsuby, zopt, zlabel, ztick, nsubz );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plbtimeCmd
 *
 * Processes plbtime Tcl command.
\*--------------------------------------------------------------------------*/

static int
plbtimeCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT year;
    PLINT month;
    PLINT day;
    PLINT hour;
    PLINT min;
    PLFLT sec;
    PLFLT ctime;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plbtime ?year month day hour min sec ctime?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 7 - 0))) ||
         (!1 && !0 && (argc != (7 + 1))) ||
         ( 1 && (argc != 1) && (argc != (7 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plbtime ?year month day hour min sec ctime?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* year is for output. */
/* month is for output. */
/* day is for output. */
/* hour is for output. */
/* min is for output. */
/* sec is for output. */
    ctime = atof(argv[1+6]);

    plbtime ( &year, &month, &day, &hour, &min, &sec, ctime );

    sprintf( buf, "%d", year );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", month );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", day );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", hour );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", min );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, sec, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+5], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plcalc_worldCmd
 *
 * Processes plcalc_world Tcl command.
\*--------------------------------------------------------------------------*/

static int
plcalc_worldCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT rx;
    PLFLT ry;
    PLFLT wx;
    PLFLT wy;
    PLINT window;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plcalc_world ?rx ry wx wy window?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 5 - 0))) ||
         (!1 && !0 && (argc != (5 + 1))) ||
         ( 1 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plcalc_world ?rx ry wx wy window?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    rx = atof(argv[1+0]);
    ry = atof(argv[1+1]);
/* wx is for output. */
/* wy is for output. */
/* window is for output. */

    plcalc_world ( rx, ry, &wx, &wy, &window );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, wx, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, wy, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", window );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plclearCmd
 *
 * Processes plclear Tcl command.
\*--------------------------------------------------------------------------*/

static int
plclearCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plclear ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plclear ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plclear (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plcol0Cmd
 *
 * Processes plcol0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plcol0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT icol0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plcol0 icol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plcol0 icol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    icol0 = atoi(argv[1+0]);

    plcol0 ( icol0 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plcol1Cmd
 *
 * Processes plcol1 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plcol1Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT col1;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plcol1 col1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plcol1 col1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    col1 = atof(argv[1+0]);

    plcol1 ( col1 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plconfigtimeCmd
 *
 * Processes plconfigtime Tcl command.
\*--------------------------------------------------------------------------*/

static int
plconfigtimeCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT scale;
    PLFLT offset1;
    PLFLT offset2;
    PLINT ccontrol;
    PLINT ifbtime_offset;
    PLINT year;
    PLINT month;
    PLINT day;
    PLINT hour;
    PLINT min;
    PLFLT sec;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plconfigtime scale offset1 offset2 ccontrol ifbtime_offset year month day hour min sec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 11 - 0))) ||
         (!0 && !0 && (argc != (11 + 1))) ||
         ( 0 && (argc != 1) && (argc != (11 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plconfigtime scale offset1 offset2 ccontrol ifbtime_offset year month day hour min sec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    scale = atof(argv[1+0]);
    offset1 = atof(argv[1+1]);
    offset2 = atof(argv[1+2]);
    ccontrol = atoi(argv[1+3]);
    ifbtime_offset = atoi(argv[1+4]);
    year = atoi(argv[1+5]);
    month = atoi(argv[1+6]);
    day = atoi(argv[1+7]);
    hour = atoi(argv[1+8]);
    min = atoi(argv[1+9]);
    sec = atof(argv[1+10]);

    plconfigtime ( scale, offset1, offset2, ccontrol, ifbtime_offset, year, month, day, hour, min, sec );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plcpstrmCmd
 *
 * Processes plcpstrm Tcl command.
\*--------------------------------------------------------------------------*/

static int
plcpstrmCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT iplsr;
    PLINT flags;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plcpstrm iplsr flags", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plcpstrm iplsr flags", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    iplsr = atoi(argv[1+0]);
    flags = atoi(argv[1+1]);

    plcpstrm ( iplsr, flags );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plctimeCmd
 *
 * Processes plctime Tcl command.
\*--------------------------------------------------------------------------*/

static int
plctimeCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT year;
    PLINT month;
    PLINT day;
    PLINT hour;
    PLINT min;
    PLFLT sec;
    PLFLT ctime;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plctime ?year month day hour min sec ctime?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 7 - 0))) ||
         (!1 && !0 && (argc != (7 + 1))) ||
         ( 1 && (argc != 1) && (argc != (7 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plctime ?year month day hour min sec ctime?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    year = atoi(argv[1+0]);
    month = atoi(argv[1+1]);
    day = atoi(argv[1+2]);
    hour = atoi(argv[1+3]);
    min = atoi(argv[1+4]);
    sec = atof(argv[1+5]);
/* ctime is for output. */

    plctime ( year, month, day, hour, min, sec, &ctime );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ctime, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+6], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pldid2pcCmd
 *
 * Processes pldid2pc Tcl command.
\*--------------------------------------------------------------------------*/

static int
pldid2pcCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT ymin;
    PLFLT xmax;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pldid2pc ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pldid2pc ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* ymin is for output. */
/* xmax is for output. */
/* ymax is for output. */

    pldid2pc ( &xmin, &ymin, &xmax, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pldip2dcCmd
 *
 * Processes pldip2dc Tcl command.
\*--------------------------------------------------------------------------*/

static int
pldip2dcCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT ymin;
    PLFLT xmax;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pldip2dc ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pldip2dc ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* ymin is for output. */
/* xmax is for output. */
/* ymax is for output. */

    pldip2dc ( &xmin, &ymin, &xmax, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plendCmd
 *
 * Processes plend Tcl command.
\*--------------------------------------------------------------------------*/

static int
plendCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plend ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plend ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plend (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plend1Cmd
 *
 * Processes plend1 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plend1Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plend1 ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plend1 ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plend1 (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plenvCmd
 *
 * Processes plenv Tcl command.
\*--------------------------------------------------------------------------*/

static int
plenvCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;
    PLINT just;
    PLINT axis;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plenv xmin xmax ymin ymax just axis", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plenv xmin xmax ymin ymax just axis", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);
    just = atoi(argv[1+4]);
    axis = atoi(argv[1+5]);

    plenv ( xmin, xmax, ymin, ymax, just, axis );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plenv0Cmd
 *
 * Processes plenv0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plenv0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;
    PLINT just;
    PLINT axis;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plenv0 xmin xmax ymin ymax just axis", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plenv0 xmin xmax ymin ymax just axis", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);
    just = atoi(argv[1+4]);
    axis = atoi(argv[1+5]);

    plenv0 ( xmin, xmax, ymin, ymax, just, axis );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pleopCmd
 *
 * Processes pleop Tcl command.
\*--------------------------------------------------------------------------*/

static int
pleopCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pleop ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pleop ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    pleop (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plerrxCmd
 *
 * Processes plerrx Tcl command.
\*--------------------------------------------------------------------------*/

static int
plerrxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *xmin;
    tclMatrix *matxmin;
    PLFLT *xmax;
    tclMatrix *matxmax;
    PLFLT *y;
    tclMatrix *maty;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plerrx n xmin xmax y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plerrx n xmin xmax y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matxmin = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matxmin == NULL) return TCL_ERROR;
    xmin = matxmin->fdata;
    matxmax = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matxmax == NULL) return TCL_ERROR;
    xmax = matxmax->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;

    plerrx ( n, xmin, xmax, y );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plerryCmd
 *
 * Processes plerry Tcl command.
\*--------------------------------------------------------------------------*/

static int
plerryCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *ymin;
    tclMatrix *matymin;
    PLFLT *ymax;
    tclMatrix *matymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plerry n x ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plerry n x ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    matymin = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matymin == NULL) return TCL_ERROR;
    ymin = matymin->fdata;
    matymax = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matymax == NULL) return TCL_ERROR;
    ymax = matymax->fdata;

    plerry ( n, x, ymin, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plfamadvCmd
 *
 * Processes plfamadv Tcl command.
\*--------------------------------------------------------------------------*/

static int
plfamadvCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plfamadv ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plfamadv ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plfamadv (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plfillCmd
 *
 * Processes plfill Tcl command.
\*--------------------------------------------------------------------------*/

static int
plfillCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plfill n x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plfill n x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;

    plfill ( n, x, y );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plfill3Cmd
 *
 * Processes plfill3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plfill3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT *z;
    tclMatrix *matz;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plfill3 n x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plfill3 n x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    matz = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matz == NULL) return TCL_ERROR;
    z = matz->fdata;

    plfill3 ( n, x, y, z );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plflushCmd
 *
 * Processes plflush Tcl command.
\*--------------------------------------------------------------------------*/

static int
plflushCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plflush ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plflush ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plflush (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plfontCmd
 *
 * Processes plfont Tcl command.
\*--------------------------------------------------------------------------*/

static int
plfontCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT ifont;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plfont ifont", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plfont ifont", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    ifont = atoi(argv[1+0]);

    plfont ( ifont );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plfontldCmd
 *
 * Processes plfontld Tcl command.
\*--------------------------------------------------------------------------*/

static int
plfontldCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT fnt;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plfontld fnt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plfontld fnt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    fnt = atoi(argv[1+0]);

    plfontld ( fnt );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgchrCmd
 *
 * Processes plgchr Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgchrCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT def;
    PLFLT ht;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgchr ?def ht?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 2 - 0))) ||
         (!1 && !0 && (argc != (2 + 1))) ||
         ( 1 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgchr ?def ht?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* def is for output. */
/* ht is for output. */

    plgchr ( &def, &ht );

    Tcl_PrintDouble( interp, def, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ht, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgcol0Cmd
 *
 * Processes plgcol0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgcol0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT icol0;
    PLINT r;
    PLINT g;
    PLINT b;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgcol0 ?icol0 r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgcol0 ?icol0 r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    icol0 = atoi(argv[1+0]);
/* r is for output. */
/* g is for output. */
/* b is for output. */

    plgcol0 ( icol0, &r, &g, &b );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", r );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", g );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", b );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgcol0aCmd
 *
 * Processes plgcol0a Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgcol0aCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT icol0;
    PLINT r;
    PLINT g;
    PLINT b;
    PLFLT a;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgcol0a ?icol0 r g b a?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 5 - 0))) ||
         (!1 && !0 && (argc != (5 + 1))) ||
         ( 1 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgcol0a ?icol0 r g b a?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    icol0 = atoi(argv[1+0]);
/* r is for output. */
/* g is for output. */
/* b is for output. */
/* a is for output. */

    plgcol0a ( icol0, &r, &g, &b, &a );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", r );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", g );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", b );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, a, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgcolbgCmd
 *
 * Processes plgcolbg Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgcolbgCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT r;
    PLINT g;
    PLINT b;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgcolbg ?r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 3 - 0))) ||
         (!1 && !0 && (argc != (3 + 1))) ||
         ( 1 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgcolbg ?r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* r is for output. */
/* g is for output. */
/* b is for output. */

    plgcolbg ( &r, &g, &b );

    sprintf( buf, "%d", r );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", g );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", b );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgcolbgaCmd
 *
 * Processes plgcolbga Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgcolbgaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT r;
    PLINT g;
    PLINT b;
    PLFLT a;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgcolbga ?r g b a?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgcolbga ?r g b a?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* r is for output. */
/* g is for output. */
/* b is for output. */
/* a is for output. */

    plgcolbga ( &r, &g, &b, &a );

    sprintf( buf, "%d", r );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", g );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", b );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, a, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgcompressionCmd
 *
 * Processes plgcompression Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgcompressionCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT compression;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgcompression ?compression?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgcompression ?compression?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* compression is for output. */

    plgcompression ( &compression );

    sprintf( buf, "%d", compression );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgdevCmd
 *
 * Processes plgdev Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgdevCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    char devnam[200];

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgdev ?devnam?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgdev ?devnam?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* devnam is for output. */

    plgdev ( devnam );

    if (argc > 1)
       Tcl_SetVar( interp, argv[1+0], devnam, 0 );
    else
        Tcl_AppendResult( interp, devnam, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgdidevCmd
 *
 * Processes plgdidev Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgdidevCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT mar;
    PLFLT aspect;
    PLFLT jx;
    PLFLT jy;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgdidev ?mar aspect jx jy?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgdidev ?mar aspect jx jy?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* mar is for output. */
/* aspect is for output. */
/* jx is for output. */
/* jy is for output. */

    plgdidev ( &mar, &aspect, &jx, &jy );

    Tcl_PrintDouble( interp, mar, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, aspect, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, jx, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, jy, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgdioriCmd
 *
 * Processes plgdiori Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgdioriCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT rot;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgdiori ?rot?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgdiori ?rot?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* rot is for output. */

    plgdiori ( &rot );

    Tcl_PrintDouble( interp, rot, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgdipltCmd
 *
 * Processes plgdiplt Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgdipltCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT ymin;
    PLFLT xmax;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgdiplt ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgdiplt ?xmin ymin xmax ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* ymin is for output. */
/* xmax is for output. */
/* ymax is for output. */

    plgdiplt ( &xmin, &ymin, &xmax, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgescCmd
 *
 * Processes plgesc Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgescCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    char esc;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgesc ?esc?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgesc ?esc?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* esc is for output. */

    plgesc ( &esc );

    sprintf( buf, "%c", esc );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgfamCmd
 *
 * Processes plgfam Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgfamCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT fam;
    PLINT num;
    PLINT bmax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgfam ?fam num bmax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 3 - 0))) ||
         (!1 && !0 && (argc != (3 + 1))) ||
         ( 1 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgfam ?fam num bmax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* fam is for output. */
/* num is for output. */
/* bmax is for output. */

    plgfam ( &fam, &num, &bmax );

    sprintf( buf, "%d", fam );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", num );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", bmax );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgfciCmd
 *
 * Processes plgfci Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgfciCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLUNICODE fci;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgfci ?fci?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgfci ?fci?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* fci is for output. */

    plgfci ( &fci );

    sprintf( buf, "%u", fci );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgfnamCmd
 *
 * Processes plgfnam Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgfnamCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    char fnam[200];

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgfnam ?fnam?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgfnam ?fnam?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* fnam is for output. */

    plgfnam ( fnam );

    if (argc > 1)
       Tcl_SetVar( interp, argv[1+0], fnam, 0 );
    else
        Tcl_AppendResult( interp, fnam, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgfontCmd
 *
 * Processes plgfont Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgfontCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT family;
    PLINT style;
    PLINT weight;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgfont ?family style weight?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 3 - 0))) ||
         (!1 && !0 && (argc != (3 + 1))) ||
         ( 1 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgfont ?family style weight?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* family is for output. */
/* style is for output. */
/* weight is for output. */

    plgfont ( &family, &style, &weight );

    sprintf( buf, "%d", family );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", style );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", weight );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plglevelCmd
 *
 * Processes plglevel Tcl command.
\*--------------------------------------------------------------------------*/

static int
plglevelCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT level;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plglevel ?level?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plglevel ?level?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* level is for output. */

    plglevel ( &level );

    sprintf( buf, "%d", level );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgpageCmd
 *
 * Processes plgpage Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgpageCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xp;
    PLFLT yp;
    PLINT xleng;
    PLINT yleng;
    PLINT xoff;
    PLINT yoff;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgpage ?xp yp xleng yleng xoff yoff?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 6 - 0))) ||
         (!1 && !0 && (argc != (6 + 1))) ||
         ( 1 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgpage ?xp yp xleng yleng xoff yoff?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xp is for output. */
/* yp is for output. */
/* xleng is for output. */
/* yleng is for output. */
/* xoff is for output. */
/* yoff is for output. */

    plgpage ( &xp, &yp, &xleng, &yleng, &xoff, &yoff );

    Tcl_PrintDouble( interp, xp, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, yp, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", xleng );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", yleng );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", xoff );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", yoff );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+5], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgraCmd
 *
 * Processes plgra Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgraCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgra ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgra ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plgra (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgradientCmd
 *
 * Processes plgradient Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgradientCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT angle;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgradient n x y angle", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgradient n x y angle", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    angle = atof(argv[1+3]);

    plgradient ( n, x, y, angle );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgspaCmd
 *
 * Processes plgspa Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgspaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgspa ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgspa ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* xmax is for output. */
/* ymin is for output. */
/* ymax is for output. */

    plgspa ( &xmin, &xmax, &ymin, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgstrmCmd
 *
 * Processes plgstrm Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgstrmCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT strm;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgstrm ?strm?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgstrm ?strm?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* strm is for output. */

    plgstrm ( &strm );

    sprintf( buf, "%d", strm );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgverCmd
 *
 * Processes plgver Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgverCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    char ver[200];

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgver ?ver?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgver ?ver?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* ver is for output. */

    plgver ( ver );

    if (argc > 1)
       Tcl_SetVar( interp, argv[1+0], ver, 0 );
    else
        Tcl_AppendResult( interp, ver, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgvpdCmd
 *
 * Processes plgvpd Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgvpdCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgvpd ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgvpd ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* xmax is for output. */
/* ymin is for output. */
/* ymax is for output. */

    plgvpd ( &xmin, &xmax, &ymin, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgvpwCmd
 *
 * Processes plgvpw Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgvpwCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgvpw ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 4 - 0))) ||
         (!1 && !0 && (argc != (4 + 1))) ||
         ( 1 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgvpw ?xmin xmax ymin ymax?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* xmin is for output. */
/* xmax is for output. */
/* ymin is for output. */
/* ymax is for output. */

    plgvpw ( &xmin, &xmax, &ymin, &ymax );

    Tcl_PrintDouble( interp, xmin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, xmax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymin, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+2], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, ymax, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgxaxCmd
 *
 * Processes plgxax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgxaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax;
    PLINT digits;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgxax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 2 - 0))) ||
         (!1 && !0 && (argc != (2 + 1))) ||
         ( 1 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgxax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* digmax is for output. */
/* digits is for output. */

    plgxax ( &digmax, &digits );

    sprintf( buf, "%d", digmax );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", digits );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgyaxCmd
 *
 * Processes plgyax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgyaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax;
    PLINT digits;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgyax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 2 - 0))) ||
         (!1 && !0 && (argc != (2 + 1))) ||
         ( 1 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgyax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* digmax is for output. */
/* digits is for output. */

    plgyax ( &digmax, &digits );

    sprintf( buf, "%d", digmax );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", digits );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plgzaxCmd
 *
 * Processes plgzax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plgzaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax;
    PLINT digits;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plgzax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 2 - 0))) ||
         (!1 && !0 && (argc != (2 + 1))) ||
         ( 1 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plgzax ?digmax digits?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* digmax is for output. */
/* digits is for output. */

    plgzax ( &digmax, &digits );

    sprintf( buf, "%d", digmax );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", digits );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plhistCmd
 *
 * Processes plhist Tcl command.
\*--------------------------------------------------------------------------*/

static int
plhistCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *data;
    tclMatrix *matdata;
    PLFLT datmin;
    PLFLT datmax;
    PLINT nbin;
    PLINT oldwin;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plhist n data datmin datmax nbin oldwin", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plhist n data datmin datmax nbin oldwin", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matdata = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matdata == NULL) return TCL_ERROR;
    data = matdata->fdata;
    datmin = atof(argv[1+2]);
    datmax = atof(argv[1+3]);
    nbin = atoi(argv[1+4]);
    oldwin = atoi(argv[1+5]);

    plhist ( n, data, datmin, datmax, nbin, oldwin );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plhlsrgbCmd
 *
 * Processes plhlsrgb Tcl command.
\*--------------------------------------------------------------------------*/

static int
plhlsrgbCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT h;
    PLFLT l;
    PLFLT s;
    PLFLT r;
    PLFLT g;
    PLFLT b;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plhlsrgb ?h l s r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 6 - 0))) ||
         (!1 && !0 && (argc != (6 + 1))) ||
         ( 1 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plhlsrgb ?h l s r g b?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    h = atof(argv[1+0]);
    l = atof(argv[1+1]);
    s = atof(argv[1+2]);
/* r is for output. */
/* g is for output. */
/* b is for output. */

    plhlsrgb ( h, l, s, &r, &g, &b );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, r, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, g, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, b, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+5], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plinitCmd
 *
 * Processes plinit Tcl command.
\*--------------------------------------------------------------------------*/

static int
plinitCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plinit ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plinit ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plinit (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pljoinCmd
 *
 * Processes pljoin Tcl command.
\*--------------------------------------------------------------------------*/

static int
pljoinCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT x1;
    PLFLT y1;
    PLFLT x2;
    PLFLT y2;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pljoin x1 y1 x2 y2", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pljoin x1 y1 x2 y2", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    x1 = atof(argv[1+0]);
    y1 = atof(argv[1+1]);
    x2 = atof(argv[1+2]);
    y2 = atof(argv[1+3]);

    pljoin ( x1, y1, x2, y2 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pllabCmd
 *
 * Processes pllab Tcl command.
\*--------------------------------------------------------------------------*/

static int
pllabCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *xlabel;
    const char *ylabel;
    const char *tlabel;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pllab xlabel ylabel tlabel", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pllab xlabel ylabel tlabel", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xlabel = argv[1+0];
    ylabel = argv[1+1];
    tlabel = argv[1+2];

    pllab ( xlabel, ylabel, tlabel );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pllightsourceCmd
 *
 * Processes pllightsource Tcl command.
\*--------------------------------------------------------------------------*/

static int
pllightsourceCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT x;
    PLFLT y;
    PLFLT z;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pllightsource x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pllightsource x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    x = atof(argv[1+0]);
    y = atof(argv[1+1]);
    z = atof(argv[1+2]);

    pllightsource ( x, y, z );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pllineCmd
 *
 * Processes plline Tcl command.
\*--------------------------------------------------------------------------*/

static int
pllineCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plline n x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plline n x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;

    plline ( n, x, y );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plline3Cmd
 *
 * Processes plline3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plline3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT *z;
    tclMatrix *matz;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plline3 n x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plline3 n x y z", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    matz = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matz == NULL) return TCL_ERROR;
    z = matz->fdata;

    plline3 ( n, x, y, z );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pllstyCmd
 *
 * Processes pllsty Tcl command.
\*--------------------------------------------------------------------------*/

static int
pllstyCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT lin;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pllsty lin", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pllsty lin", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    lin = atoi(argv[1+0]);

    pllsty ( lin );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plmkstrmCmd
 *
 * Processes plmkstrm Tcl command.
\*--------------------------------------------------------------------------*/

static int
plmkstrmCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT strm;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plmkstrm ?strm?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 1 - 0))) ||
         (!1 && !0 && (argc != (1 + 1))) ||
         ( 1 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plmkstrm ?strm?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

/* strm is for output. */

    plmkstrm ( &strm );

    sprintf( buf, "%d", strm );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+0], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plmtexCmd
 *
 * Processes plmtex Tcl command.
\*--------------------------------------------------------------------------*/

static int
plmtexCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *side;
    PLFLT disp;
    PLFLT pos;
    PLFLT just;
    const char *text;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plmtex side disp pos just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plmtex side disp pos just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    side = argv[1+0];
    disp = atof(argv[1+1]);
    pos = atof(argv[1+2]);
    just = atof(argv[1+3]);
    text = argv[1+4];

    plmtex ( side, disp, pos, just, text );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plmtex3Cmd
 *
 * Processes plmtex3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plmtex3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *side;
    PLFLT disp;
    PLFLT pos;
    PLFLT just;
    const char *text;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plmtex3 side disp pos just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plmtex3 side disp pos just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    side = argv[1+0];
    disp = atof(argv[1+1]);
    pos = atof(argv[1+2]);
    just = atof(argv[1+3]);
    text = argv[1+4];

    plmtex3 ( side, disp, pos, just, text );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plpatCmd
 *
 * Processes plpat Tcl command.
\*--------------------------------------------------------------------------*/

static int
plpatCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT nlin;
    PLINT *inc;
    tclMatrix *matinc;
    PLINT *del;
    tclMatrix *matdel;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plpat nlin inc del", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plpat nlin inc del", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    nlin = atoi(argv[1+0]);
    matinc = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matinc == NULL) return TCL_ERROR;
    inc = matinc->idata;
    matdel = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matdel == NULL) return TCL_ERROR;
    del = matdel->idata;

    plpat ( nlin, inc, del );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plpoinCmd
 *
 * Processes plpoin Tcl command.
\*--------------------------------------------------------------------------*/

static int
plpoinCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLINT code;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plpoin n x y code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plpoin n x y code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    code = atoi(argv[1+3]);

    plpoin ( n, x, y, code );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plpoin3Cmd
 *
 * Processes plpoin3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plpoin3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT *z;
    tclMatrix *matz;
    PLINT code;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plpoin3 n x y z code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plpoin3 n x y z code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    matz = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matz == NULL) return TCL_ERROR;
    z = matz->fdata;
    code = atoi(argv[1+4]);

    plpoin3 ( n, x, y, z, code );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plpoly3Cmd
 *
 * Processes plpoly3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plpoly3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT *z;
    tclMatrix *matz;
    PLINT *draw;
    tclMatrix *matdraw;
    PLINT ifcc;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plpoly3 n x y z draw ifcc", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plpoly3 n x y z draw ifcc", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    matz = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matz == NULL) return TCL_ERROR;
    z = matz->fdata;
    matdraw = Tcl_GetMatrixPtr( interp, argv[1+4] );
    if (matdraw == NULL) return TCL_ERROR;
    draw = matdraw->idata;
    ifcc = atoi(argv[1+5]);

    plpoly3 ( n, x, y, z, draw, ifcc );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plprecCmd
 *
 * Processes plprec Tcl command.
\*--------------------------------------------------------------------------*/

static int
plprecCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT setp;
    PLINT prec;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plprec setp prec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plprec setp prec", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    setp = atoi(argv[1+0]);
    prec = atoi(argv[1+1]);

    plprec ( setp, prec );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plpstyCmd
 *
 * Processes plpsty Tcl command.
\*--------------------------------------------------------------------------*/

static int
plpstyCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT patt;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plpsty patt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plpsty patt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    patt = atoi(argv[1+0]);

    plpsty ( patt );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plptexCmd
 *
 * Processes plptex Tcl command.
\*--------------------------------------------------------------------------*/

static int
plptexCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT x;
    PLFLT y;
    PLFLT dx;
    PLFLT dy;
    PLFLT just;
    const char *text;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plptex x y dx dy just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plptex x y dx dy just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    x = atof(argv[1+0]);
    y = atof(argv[1+1]);
    dx = atof(argv[1+2]);
    dy = atof(argv[1+3]);
    just = atof(argv[1+4]);
    text = argv[1+5];

    plptex ( x, y, dx, dy, just, text );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plptex3Cmd
 *
 * Processes plptex3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plptex3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT wx;
    PLFLT wy;
    PLFLT wz;
    PLFLT dx;
    PLFLT dy;
    PLFLT dz;
    PLFLT sx;
    PLFLT sy;
    PLFLT sz;
    PLFLT just;
    const char *text;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plptex3 wx wy wz dx dy dz sx sy sz just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 11 - 0))) ||
         (!0 && !0 && (argc != (11 + 1))) ||
         ( 0 && (argc != 1) && (argc != (11 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plptex3 wx wy wz dx dy dz sx sy sz just text", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    wx = atof(argv[1+0]);
    wy = atof(argv[1+1]);
    wz = atof(argv[1+2]);
    dx = atof(argv[1+3]);
    dy = atof(argv[1+4]);
    dz = atof(argv[1+5]);
    sx = atof(argv[1+6]);
    sy = atof(argv[1+7]);
    sz = atof(argv[1+8]);
    just = atof(argv[1+9]);
    text = argv[1+10];

    plptex3 ( wx, wy, wz, dx, dy, dz, sx, sy, sz, just, text );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plreplotCmd
 *
 * Processes plreplot Tcl command.
\*--------------------------------------------------------------------------*/

static int
plreplotCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plreplot ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plreplot ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plreplot (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plrgbhlsCmd
 *
 * Processes plrgbhls Tcl command.
\*--------------------------------------------------------------------------*/

static int
plrgbhlsCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT r;
    PLFLT g;
    PLFLT b;
    PLFLT h;
    PLFLT l;
    PLFLT s;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plrgbhls ?r g b h l s?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 6 - 0))) ||
         (!1 && !0 && (argc != (6 + 1))) ||
         ( 1 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plrgbhls ?r g b h l s?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    r = atof(argv[1+0]);
    g = atof(argv[1+1]);
    b = atof(argv[1+2]);
/* h is for output. */
/* l is for output. */
/* s is for output. */

    plrgbhls ( r, g, b, &h, &l, &s );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, h, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+3], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, l, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+4], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );
    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    Tcl_PrintDouble( interp, s, buf );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+5], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plschrCmd
 *
 * Processes plschr Tcl command.
\*--------------------------------------------------------------------------*/

static int
plschrCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT def;
    PLFLT scale;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plschr def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plschr def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    def = atof(argv[1+0]);
    scale = atof(argv[1+1]);

    plschr ( def, scale );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap0Cmd
 *
 * Processes plscmap0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT *r;
    tclMatrix *matr;
    PLINT *g;
    tclMatrix *matg;
    PLINT *b;
    tclMatrix *matb;
    PLINT ncol0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap0 r g b ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap0 r g b ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    matr = Tcl_GetMatrixPtr( interp, argv[1+0] );
    if (matr == NULL) return TCL_ERROR;
    r = matr->idata;
    matg = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matg == NULL) return TCL_ERROR;
    g = matg->idata;
    matb = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matb == NULL) return TCL_ERROR;
    b = matb->idata;
    ncol0 = atoi(argv[1+3]);

    plscmap0 ( r, g, b, ncol0 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap0aCmd
 *
 * Processes plscmap0a Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap0aCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT *r;
    tclMatrix *matr;
    PLINT *g;
    tclMatrix *matg;
    PLINT *b;
    tclMatrix *matb;
    PLFLT *a;
    tclMatrix *mata;
    PLINT ncol0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap0a r g b a ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap0a r g b a ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    matr = Tcl_GetMatrixPtr( interp, argv[1+0] );
    if (matr == NULL) return TCL_ERROR;
    r = matr->idata;
    matg = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matg == NULL) return TCL_ERROR;
    g = matg->idata;
    matb = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matb == NULL) return TCL_ERROR;
    b = matb->idata;
    mata = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (mata == NULL) return TCL_ERROR;
    a = mata->fdata;
    ncol0 = atoi(argv[1+4]);

    plscmap0a ( r, g, b, a, ncol0 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap0nCmd
 *
 * Processes plscmap0n Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap0nCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT ncol0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap0n ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap0n ncol0", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    ncol0 = atoi(argv[1+0]);

    plscmap0n ( ncol0 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap1Cmd
 *
 * Processes plscmap1 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap1Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT *r;
    tclMatrix *matr;
    PLINT *g;
    tclMatrix *matg;
    PLINT *b;
    tclMatrix *matb;
    PLINT ncol1;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap1 r g b ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap1 r g b ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    matr = Tcl_GetMatrixPtr( interp, argv[1+0] );
    if (matr == NULL) return TCL_ERROR;
    r = matr->idata;
    matg = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matg == NULL) return TCL_ERROR;
    g = matg->idata;
    matb = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matb == NULL) return TCL_ERROR;
    b = matb->idata;
    ncol1 = atoi(argv[1+3]);

    plscmap1 ( r, g, b, ncol1 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap1aCmd
 *
 * Processes plscmap1a Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap1aCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT *r;
    tclMatrix *matr;
    PLINT *g;
    tclMatrix *matg;
    PLINT *b;
    tclMatrix *matb;
    PLFLT *a;
    tclMatrix *mata;
    PLINT ncol1;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap1a r g b a ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap1a r g b a ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    matr = Tcl_GetMatrixPtr( interp, argv[1+0] );
    if (matr == NULL) return TCL_ERROR;
    r = matr->idata;
    matg = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matg == NULL) return TCL_ERROR;
    g = matg->idata;
    matb = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matb == NULL) return TCL_ERROR;
    b = matb->idata;
    mata = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (mata == NULL) return TCL_ERROR;
    a = mata->fdata;
    ncol1 = atoi(argv[1+4]);

    plscmap1a ( r, g, b, a, ncol1 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap1lCmd
 *
 * Processes plscmap1l Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap1lCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT itype;
    PLINT npts;
    PLFLT *intensity;
    tclMatrix *matintensity;
    PLFLT *coord1;
    tclMatrix *matcoord1;
    PLFLT *coord2;
    tclMatrix *matcoord2;
    PLFLT *coord3;
    tclMatrix *matcoord3;
    PLINT *alt_hue_path;
    tclMatrix *matalt_hue_path;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap1l itype npts intensity coord1 coord2 coord3 alt_hue_path", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 7 - 0))) ||
         (!0 && !0 && (argc != (7 + 1))) ||
         ( 0 && (argc != 1) && (argc != (7 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap1l itype npts intensity coord1 coord2 coord3 alt_hue_path", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    itype = atoi(argv[1+0]);
    npts = atoi(argv[1+1]);
    matintensity = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matintensity == NULL) return TCL_ERROR;
    intensity = matintensity->fdata;
    matcoord1 = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matcoord1 == NULL) return TCL_ERROR;
    coord1 = matcoord1->fdata;
    matcoord2 = Tcl_GetMatrixPtr( interp, argv[1+4] );
    if (matcoord2 == NULL) return TCL_ERROR;
    coord2 = matcoord2->fdata;
    matcoord3 = Tcl_GetMatrixPtr( interp, argv[1+5] );
    if (matcoord3 == NULL) return TCL_ERROR;
    coord3 = matcoord3->fdata;
    matalt_hue_path = Tcl_GetMatrixPtr( interp, argv[1+6] );
    if (matalt_hue_path == NULL) return TCL_ERROR;
    alt_hue_path = matalt_hue_path->idata;

    plscmap1l ( itype, npts, intensity, coord1, coord2, coord3, alt_hue_path );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap1laCmd
 *
 * Processes plscmap1la Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap1laCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT itype;
    PLINT npts;
    PLFLT *intensity;
    tclMatrix *matintensity;
    PLFLT *coord1;
    tclMatrix *matcoord1;
    PLFLT *coord2;
    tclMatrix *matcoord2;
    PLFLT *coord3;
    tclMatrix *matcoord3;
    PLFLT *a;
    tclMatrix *mata;
    PLINT *alt_hue_path;
    tclMatrix *matalt_hue_path;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap1la itype npts intensity coord1 coord2 coord3 a alt_hue_path", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 8 - 0))) ||
         (!0 && !0 && (argc != (8 + 1))) ||
         ( 0 && (argc != 1) && (argc != (8 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap1la itype npts intensity coord1 coord2 coord3 a alt_hue_path", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    itype = atoi(argv[1+0]);
    npts = atoi(argv[1+1]);
    matintensity = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matintensity == NULL) return TCL_ERROR;
    intensity = matintensity->fdata;
    matcoord1 = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matcoord1 == NULL) return TCL_ERROR;
    coord1 = matcoord1->fdata;
    matcoord2 = Tcl_GetMatrixPtr( interp, argv[1+4] );
    if (matcoord2 == NULL) return TCL_ERROR;
    coord2 = matcoord2->fdata;
    matcoord3 = Tcl_GetMatrixPtr( interp, argv[1+5] );
    if (matcoord3 == NULL) return TCL_ERROR;
    coord3 = matcoord3->fdata;
    mata = Tcl_GetMatrixPtr( interp, argv[1+6] );
    if (mata == NULL) return TCL_ERROR;
    a = mata->fdata;
    matalt_hue_path = Tcl_GetMatrixPtr( interp, argv[1+7] );
    if (matalt_hue_path == NULL) return TCL_ERROR;
    alt_hue_path = matalt_hue_path->idata;

    plscmap1la ( itype, npts, intensity, coord1, coord2, coord3, a, alt_hue_path );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscmap1nCmd
 *
 * Processes plscmap1n Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscmap1nCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT ncol1;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscmap1n ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscmap1n ncol1", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    ncol1 = atoi(argv[1+0]);

    plscmap1n ( ncol1 );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscol0Cmd
 *
 * Processes plscol0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscol0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT icol0;
    PLINT r;
    PLINT g;
    PLINT b;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscol0 icol0 r g b", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscol0 icol0 r g b", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    icol0 = atoi(argv[1+0]);
    r = atoi(argv[1+1]);
    g = atoi(argv[1+2]);
    b = atoi(argv[1+3]);

    plscol0 ( icol0, r, g, b );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscol0aCmd
 *
 * Processes plscol0a Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscol0aCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT icol0;
    PLINT r;
    PLINT g;
    PLINT b;
    PLFLT a;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscol0a icol0 r g b a", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscol0a icol0 r g b a", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    icol0 = atoi(argv[1+0]);
    r = atoi(argv[1+1]);
    g = atoi(argv[1+2]);
    b = atoi(argv[1+3]);
    a = atof(argv[1+4]);

    plscol0a ( icol0, r, g, b, a );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscolbgCmd
 *
 * Processes plscolbg Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscolbgCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT r;
    PLINT g;
    PLINT b;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscolbg r g b", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscolbg r g b", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    r = atoi(argv[1+0]);
    g = atoi(argv[1+1]);
    b = atoi(argv[1+2]);

    plscolbg ( r, g, b );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscolbgaCmd
 *
 * Processes plscolbga Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscolbgaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT r;
    PLINT g;
    PLINT b;
    PLFLT a;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscolbga r g b a", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscolbga r g b a", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    r = atoi(argv[1+0]);
    g = atoi(argv[1+1]);
    b = atoi(argv[1+2]);
    a = atof(argv[1+3]);

    plscolbga ( r, g, b, a );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscolorCmd
 *
 * Processes plscolor Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscolorCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT color;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscolor color", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscolor color", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    color = atoi(argv[1+0]);

    plscolor ( color );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plscompressionCmd
 *
 * Processes plscompression Tcl command.
\*--------------------------------------------------------------------------*/

static int
plscompressionCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT compression;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plscompression compression", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plscompression compression", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    compression = atoi(argv[1+0]);

    plscompression ( compression );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdevCmd
 *
 * Processes plsdev Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdevCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *devnam;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdev devnam", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdev devnam", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    devnam = argv[1+0];

    plsdev ( devnam );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdidevCmd
 *
 * Processes plsdidev Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdidevCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT mar;
    PLFLT aspect;
    PLFLT jx;
    PLFLT jy;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdidev mar aspect jx jy", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdidev mar aspect jx jy", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    mar = atof(argv[1+0]);
    aspect = atof(argv[1+1]);
    jx = atof(argv[1+2]);
    jy = atof(argv[1+3]);

    plsdidev ( mar, aspect, jx, jy );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdimapCmd
 *
 * Processes plsdimap Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdimapCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT dimxmin;
    PLINT dimxmax;
    PLINT dimymin;
    PLINT dimymax;
    PLFLT dimxpmm;
    PLFLT dimypmm;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdimap dimxmin dimxmax dimymin dimymax dimxpmm dimypmm", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdimap dimxmin dimxmax dimymin dimymax dimxpmm dimypmm", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    dimxmin = atoi(argv[1+0]);
    dimxmax = atoi(argv[1+1]);
    dimymin = atoi(argv[1+2]);
    dimymax = atoi(argv[1+3]);
    dimxpmm = atof(argv[1+4]);
    dimypmm = atof(argv[1+5]);

    plsdimap ( dimxmin, dimxmax, dimymin, dimymax, dimxpmm, dimypmm );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdioriCmd
 *
 * Processes plsdiori Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdioriCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT rot;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdiori rot", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdiori rot", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    rot = atof(argv[1+0]);

    plsdiori ( rot );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdipltCmd
 *
 * Processes plsdiplt Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdipltCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT ymin;
    PLFLT xmax;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdiplt xmin ymin xmax ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdiplt xmin ymin xmax ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    ymin = atof(argv[1+1]);
    xmax = atof(argv[1+2]);
    ymax = atof(argv[1+3]);

    plsdiplt ( xmin, ymin, xmax, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsdiplzCmd
 *
 * Processes plsdiplz Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsdiplzCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT ymin;
    PLFLT xmax;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsdiplz xmin ymin xmax ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsdiplz xmin ymin xmax ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    ymin = atof(argv[1+1]);
    xmax = atof(argv[1+2]);
    ymax = atof(argv[1+3]);

    plsdiplz ( xmin, ymin, xmax, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plseedCmd
 *
 * Processes plseed Tcl command.
\*--------------------------------------------------------------------------*/

static int
plseedCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    unsigned int s;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plseed s", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plseed s", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    s = (unsigned int) strtoul(argv[1+0],NULL,10);

    plseed ( s );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsescCmd
 *
 * Processes plsesc Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsescCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    char esc;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsesc esc", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsesc esc", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    esc = argv[1+0][0];

    plsesc ( esc );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsfamCmd
 *
 * Processes plsfam Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsfamCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT fam;
    PLINT num;
    PLINT bmax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsfam fam num bmax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsfam fam num bmax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    fam = atoi(argv[1+0]);
    num = atoi(argv[1+1]);
    bmax = atoi(argv[1+2]);

    plsfam ( fam, num, bmax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsfciCmd
 *
 * Processes plsfci Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsfciCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLUNICODE fci;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsfci fci", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsfci fci", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    fci = (PLUNICODE) strtoul(argv[1+0],NULL,10);

    plsfci ( fci );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsfnamCmd
 *
 * Processes plsfnam Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsfnamCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *fnam;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsfnam fnam", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsfnam fnam", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    fnam = argv[1+0];

    plsfnam ( fnam );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsfontCmd
 *
 * Processes plsfont Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsfontCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT family;
    PLINT style;
    PLINT weight;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsfont family style weight", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsfont family style weight", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    family = atoi(argv[1+0]);
    style = atoi(argv[1+1]);
    weight = atoi(argv[1+2]);

    plsfont ( family, style, weight );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsmajCmd
 *
 * Processes plsmaj Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsmajCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT def;
    PLFLT scale;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsmaj def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsmaj def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    def = atof(argv[1+0]);
    scale = atof(argv[1+1]);

    plsmaj ( def, scale );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsminCmd
 *
 * Processes plsmin Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsminCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT def;
    PLFLT scale;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsmin def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsmin def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    def = atof(argv[1+0]);
    scale = atof(argv[1+1]);

    plsmin ( def, scale );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsoriCmd
 *
 * Processes plsori Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsoriCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT ori;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsori ori", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsori ori", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    ori = atoi(argv[1+0]);

    plsori ( ori );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plspageCmd
 *
 * Processes plspage Tcl command.
\*--------------------------------------------------------------------------*/

static int
plspageCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xp;
    PLFLT yp;
    PLINT xleng;
    PLINT yleng;
    PLINT xoff;
    PLINT yoff;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plspage xp yp xleng yleng xoff yoff", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 6 - 0))) ||
         (!0 && !0 && (argc != (6 + 1))) ||
         ( 0 && (argc != 1) && (argc != (6 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plspage xp yp xleng yleng xoff yoff", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xp = atof(argv[1+0]);
    yp = atof(argv[1+1]);
    xleng = atoi(argv[1+2]);
    yleng = atoi(argv[1+3]);
    xoff = atoi(argv[1+4]);
    yoff = atoi(argv[1+5]);

    plspage ( xp, yp, xleng, yleng, xoff, yoff );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plspal0Cmd
 *
 * Processes plspal0 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plspal0Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *filename;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plspal0 filename", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plspal0 filename", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    filename = argv[1+0];

    plspal0 ( filename );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plspal1Cmd
 *
 * Processes plspal1 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plspal1Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *filename;
    PLINT interpolate;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plspal1 filename interpolate", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plspal1 filename interpolate", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    filename = argv[1+0];
    interpolate = atoi(argv[1+1]);

    plspal1 ( filename, interpolate );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plspauseCmd
 *
 * Processes plspause Tcl command.
\*--------------------------------------------------------------------------*/

static int
plspauseCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT paus;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plspause paus", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plspause paus", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    paus = atoi(argv[1+0]);

    plspause ( paus );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsstrmCmd
 *
 * Processes plsstrm Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsstrmCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT strm;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsstrm strm", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsstrm strm", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    strm = atoi(argv[1+0]);

    plsstrm ( strm );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plssubCmd
 *
 * Processes plssub Tcl command.
\*--------------------------------------------------------------------------*/

static int
plssubCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT nx;
    PLINT ny;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plssub nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plssub nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    nx = atoi(argv[1+0]);
    ny = atoi(argv[1+1]);

    plssub ( nx, ny );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plssymCmd
 *
 * Processes plssym Tcl command.
\*--------------------------------------------------------------------------*/

static int
plssymCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT def;
    PLFLT scale;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plssym def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plssym def scale", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    def = atof(argv[1+0]);
    scale = atof(argv[1+1]);

    plssym ( def, scale );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstarCmd
 *
 * Processes plstar Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstarCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT nx;
    PLINT ny;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstar nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstar nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    nx = atoi(argv[1+0]);
    ny = atoi(argv[1+1]);

    plstar ( nx, ny );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstartCmd
 *
 * Processes plstart Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstartCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *devname;
    PLINT nx;
    PLINT ny;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstart devname nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstart devname nx ny", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    devname = argv[1+0];
    nx = atoi(argv[1+1]);
    ny = atoi(argv[1+2]);

    plstart ( devname, nx, ny );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstringCmd
 *
 * Processes plstring Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstringCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    const char *string;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstring n x y string", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstring n x y string", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    string = argv[1+3];

    plstring ( n, x, y, string );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstring3Cmd
 *
 * Processes plstring3 Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstring3Cmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLFLT *z;
    tclMatrix *matz;
    const char *string;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstring3 n x y z string", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstring3 n x y z string", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    matz = Tcl_GetMatrixPtr( interp, argv[1+3] );
    if (matz == NULL) return TCL_ERROR;
    z = matz->fdata;
    string = argv[1+4];

    plstring3 ( n, x, y, z, string );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstripaCmd
 *
 * Processes plstripa Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstripaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT id;
    PLINT pen;
    PLFLT x;
    PLFLT y;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstripa id pen x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstripa id pen x y", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    id = atoi(argv[1+0]);
    pen = atoi(argv[1+1]);
    x = atof(argv[1+2]);
    y = atof(argv[1+3]);

    plstripa ( id, pen, x, y );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstripdCmd
 *
 * Processes plstripd Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstripdCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT id;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstripd id", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstripd id", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    id = atoi(argv[1+0]);

    plstripd ( id );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plstylCmd
 *
 * Processes plstyl Tcl command.
\*--------------------------------------------------------------------------*/

static int
plstylCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT nms;
    PLINT *mark;
    tclMatrix *matmark;
    PLINT *space;
    tclMatrix *matspace;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plstyl nms mark space", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 3 - 0))) ||
         (!0 && !0 && (argc != (3 + 1))) ||
         ( 0 && (argc != 1) && (argc != (3 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plstyl nms mark space", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    nms = atoi(argv[1+0]);
    matmark = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matmark == NULL) return TCL_ERROR;
    mark = matmark->idata;
    matspace = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (matspace == NULL) return TCL_ERROR;
    space = matspace->idata;

    plstyl ( nms, mark, space );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsvectCmd
 *
 * Processes plsvect Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsvectCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT *arrowx;
    tclMatrix *matarrowx;
    PLFLT *arrowy;
    tclMatrix *matarrowy;
    PLINT npts;
    PLINT fill;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsvect arrowx arrowy npts fill", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsvect arrowx arrowy npts fill", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    matarrowx = Tcl_GetMatrixPtr( interp, argv[1+0] );
    if (matarrowx == NULL) return TCL_ERROR;
    arrowx = matarrowx->fdata;
    matarrowy = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matarrowy == NULL) return TCL_ERROR;
    arrowy = matarrowy->fdata;
    npts = atoi(argv[1+2]);
    fill = atoi(argv[1+3]);

    plsvect ( arrowx, arrowy, npts, fill );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsvpaCmd
 *
 * Processes plsvpa Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsvpaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsvpa xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsvpa xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);

    plsvpa ( xmin, xmax, ymin, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsxaxCmd
 *
 * Processes plsxax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsxaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax;
    PLINT digits;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsxax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsxax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    digmax = atoi(argv[1+0]);
    digits = atoi(argv[1+1]);

    plsxax ( digmax, digits );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsyaxCmd
 *
 * Processes plsyax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsyaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax = 0;
    PLINT digits = 0;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsyax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 2 && (argc < (1 + 2 - 2))) ||
         (!0 && !2 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsyax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if (argc > 0+1) {    
    digmax = atoi(argv[1+0]);
    }
    if (argc > 1+1) {    
    digits = atoi(argv[1+1]);
    }

    plsyax ( digmax, digits );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plsymCmd
 *
 * Processes plsym Tcl command.
\*--------------------------------------------------------------------------*/

static int
plsymCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT n;
    PLFLT *x;
    tclMatrix *matx;
    PLFLT *y;
    tclMatrix *maty;
    PLINT code;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plsym n x y code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plsym n x y code", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    n = atoi(argv[1+0]);
    matx = Tcl_GetMatrixPtr( interp, argv[1+1] );
    if (matx == NULL) return TCL_ERROR;
    x = matx->fdata;
    maty = Tcl_GetMatrixPtr( interp, argv[1+2] );
    if (maty == NULL) return TCL_ERROR;
    y = maty->fdata;
    code = atoi(argv[1+3]);

    plsym ( n, x, y, code );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plszaxCmd
 *
 * Processes plszax Tcl command.
\*--------------------------------------------------------------------------*/

static int
plszaxCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT digmax;
    PLINT digits;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plszax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 2 - 0))) ||
         (!0 && !0 && (argc != (2 + 1))) ||
         ( 0 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plszax digmax digits", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    digmax = atoi(argv[1+0]);
    digits = atoi(argv[1+1]);

    plszax ( digmax, digits );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pltextCmd
 *
 * Processes pltext Tcl command.
\*--------------------------------------------------------------------------*/

static int
pltextCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pltext ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pltext ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    pltext (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * pltimefmtCmd
 *
 * Processes pltimefmt Tcl command.
\*--------------------------------------------------------------------------*/

static int
pltimefmtCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    const char *fmt;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "pltimefmt fmt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "pltimefmt fmt", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    fmt = argv[1+0];

    pltimefmt ( fmt );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plvaspCmd
 *
 * Processes plvasp Tcl command.
\*--------------------------------------------------------------------------*/

static int
plvaspCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT aspect;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plvasp aspect", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plvasp aspect", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    aspect = atof(argv[1+0]);

    plvasp ( aspect );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plvpasCmd
 *
 * Processes plvpas Tcl command.
\*--------------------------------------------------------------------------*/

static int
plvpasCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;
    PLFLT aspect;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plvpas xmin xmax ymin ymax aspect", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 5 - 0))) ||
         (!0 && !0 && (argc != (5 + 1))) ||
         ( 0 && (argc != 1) && (argc != (5 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plvpas xmin xmax ymin ymax aspect", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);
    aspect = atof(argv[1+4]);

    plvpas ( xmin, xmax, ymin, ymax, aspect );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plvporCmd
 *
 * Processes plvpor Tcl command.
\*--------------------------------------------------------------------------*/

static int
plvporCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plvpor xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plvpor xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);

    plvpor ( xmin, xmax, ymin, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plvstaCmd
 *
 * Processes plvsta Tcl command.
\*--------------------------------------------------------------------------*/

static int
plvstaCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plvsta ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 0 - 0))) ||
         (!0 && !0 && (argc != (0 + 1))) ||
         ( 0 && (argc != 1) && (argc != (0 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plvsta ", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }


    plvsta (  );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plw3dCmd
 *
 * Processes plw3d Tcl command.
\*--------------------------------------------------------------------------*/

static int
plw3dCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT basex;
    PLFLT basey;
    PLFLT height;
    PLFLT xmin0;
    PLFLT xmax0;
    PLFLT ymin0;
    PLFLT ymax0;
    PLFLT zmin0;
    PLFLT zmax0;
    PLFLT alt;
    PLFLT az;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plw3d basex basey height xmin0 xmax0 ymin0 ymax0 zmin0 zmax0 alt az", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 11 - 0))) ||
         (!0 && !0 && (argc != (11 + 1))) ||
         ( 0 && (argc != 1) && (argc != (11 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plw3d basex basey height xmin0 xmax0 ymin0 ymax0 zmin0 zmax0 alt az", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    basex = atof(argv[1+0]);
    basey = atof(argv[1+1]);
    height = atof(argv[1+2]);
    xmin0 = atof(argv[1+3]);
    xmax0 = atof(argv[1+4]);
    ymin0 = atof(argv[1+5]);
    ymax0 = atof(argv[1+6]);
    zmin0 = atof(argv[1+7]);
    zmax0 = atof(argv[1+8]);
    alt = atof(argv[1+9]);
    az = atof(argv[1+10]);

    plw3d ( basex, basey, height, xmin0, xmax0, ymin0, ymax0, zmin0, zmax0, alt, az );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plwidCmd
 *
 * Processes plwid Tcl command.
\*--------------------------------------------------------------------------*/

static int
plwidCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT width;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plwid width", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 1 - 0))) ||
         (!0 && !0 && (argc != (1 + 1))) ||
         ( 0 && (argc != 1) && (argc != (1 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plwid width", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    width = atoi(argv[1+0]);

    plwid ( width );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plwindCmd
 *
 * Processes plwind Tcl command.
\*--------------------------------------------------------------------------*/

static int
plwindCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLFLT xmin;
    PLFLT xmax;
    PLFLT ymin;
    PLFLT ymax;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plwind xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!0 && 0 && (argc < (1 + 4 - 0))) ||
         (!0 && !0 && (argc != (4 + 1))) ||
         ( 0 && (argc != 1) && (argc != (4 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plwind xmin xmax ymin ymax", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    xmin = atof(argv[1+0]);
    xmax = atof(argv[1+1]);
    ymin = atof(argv[1+2]);
    ymax = atof(argv[1+3]);

    plwind ( xmin, xmax, ymin, ymax );


    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

/*--------------------------------------------------------------------------*\
 * plxormodCmd
 *
 * Processes plxormod Tcl command.
\*--------------------------------------------------------------------------*/

static int
plxormodCmd( ClientData PL_UNUSED( clientData ), Tcl_Interp *interp, int argc, const char *argv[] )
{
    PLINT mod;
    PLINT st;

    pl_errcode = 0; errmsg[0] = '\0';

    if ( (argc == 2) && (strlen(argv[1])>0) && (strncmp(argv[1],"-help",strlen(argv[1])) == 0) ) {
	Tcl_AppendResult( interp, "command syntax: \"",
			  "plxormod ?mod st?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    if ( (!1 && 0 && (argc < (1 + 2 - 0))) ||
         (!1 && !0 && (argc != (2 + 1))) ||
         ( 1 && (argc != 1) && (argc != (2 + 1))) ) {
	Tcl_AppendResult( interp, "wrong # args: should be \"",
			  "plxormod ?mod st?", "\"",
			  (char *) NULL);
	return TCL_ERROR;
    }

    mod = atoi(argv[1+0]);
/* st is for output. */

    plxormod ( mod, &st );

    if (argc == 1)
        Tcl_AppendResult( interp, " ", (char *) NULL );
    sprintf( buf, "%d", st );
    if (argc > 1)
        Tcl_SetVar( interp, argv[1+1], buf, 0 );
    else
        Tcl_AppendResult( interp, buf, (char *) NULL );

    if (pl_errcode != 0) {
	Tcl_AppendResult(interp, errmsg, (char *) NULL);
	return TCL_ERROR;
    }

    plflush();
    return TCL_OK;
}

